# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

import sphinx_bootstrap_theme

import pywreck

# -- Project information -----------------------------------------------------

project = "pywreck"
copyright = "2020, Allan Feldman"
author = "Allan Feldman"

# The short X.Y version
version = ".".join(pywreck.__version__.split(".")[:2])
# The full version, including alpha/beta/rc tags
release = pywreck.__version__


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.viewcode",
    "sphinx.ext.githubpages",
    "sphinx.ext.intersphinx",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "bootstrap"
html_theme_path = sphinx_bootstrap_theme.get_html_theme_path()

html_theme_options = {
    "navbar_pagenav": False,
    "globaltoc_includehidden": "false",
}

html_show_sphinx = False

# -- Extension configuration -------------------------------------------------
autodoc_default_options = {
    "members": True,
    "member-order": "bysource",
}

# -- Options for intersphinx extension ---------------------------------------

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    "https://docs.python.org/3/": None,
}



# Copyright 2020 Allan Feldman
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import functools
import socket

import pytest

import pywreck

from .handlers import (
    handle_chunked,
    handle_close,
    handle_cookies,
    handle_echo,
    handle_multi_response_headers,
)


@pytest.fixture(scope="session")
def loop():
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    return loop


@pytest.fixture
def handler(request):
    return request.param


@pytest.fixture
def port(loop, handler):
    # Get a random port
    s = socket.socket()
    s.bind(("", 0))
    _port = s.getsockname()[1]
    s.close()

    server = loop.run_until_complete(asyncio.start_server(handler, "127.0.0.1", _port))
    yield _port

    async def shutdown():
        server.close()
        loop.create_task(server.wait_closed())

        tasks = asyncio.all_tasks()
        tasks.remove(shutdown_task)
        await asyncio.wait_for(
            asyncio.gather(*tasks),
            timeout=0.1,
        )

    shutdown_task = loop.create_task(shutdown())
    loop.run_until_complete(shutdown_task)


@pytest.mark.parametrize(
    "method",
    ("get", "head", "post", "put", "delete"),
)
@pytest.mark.parametrize("handler", (handle_echo,), indirect=True)
def test_basic(loop, port, method):
    response = loop.run_until_complete(
        getattr(pywreck, method)(
            "localhost",
            "/",
            headers={"user-agent": "pywreck test, yo!"},
            port=port,
            ssl=False,
            timeout=0.2,
        )
    )
    assert response.status == 200

    data = response.data
    expected_data = (
        f"{method.upper()} / HTTP/1.1\r\n"
        "host:localhost\r\n"
        "user-agent:pywreck test, yo!\r\n"
        "\r\n"
    ).encode("utf-8")

    assert response.headers == {"content-length": str(len(expected_data))}
    if method == "head":
        assert not data
    else:
        assert data == expected_data


@pytest.mark.parametrize("handler", (handle_echo,), indirect=True)
def test_multiple_requests_on_a_single_connection(loop, port):
    def validate_response(response):
        assert response.status == 200
        data = response.data
        expected_data = (
            "GET / HTTP/1.1\r\n"
            "host:localhost\r\n"
            "user-agent:pywreck test, yo!\r\n"
            "\r\n"
        ).encode("utf-8")

        assert response.headers == {"content-length": str(len(expected_data))}
        assert data == expected_data

    async def _async():
        connection = await pywreck.Connection.create(
            "localhost",
            port=port,
            ssl=False,
            timeout=None,
        )

        async with connection:
            for _ in range(2):
                response = await connection.request(
                    "GET",
                    "/",
                    headers={"user-agent": "pywreck test, yo!"},
                )
                validate_response(response)

    loop.run_until_complete(_async())


@pytest.mark.parametrize("handler", (handle_echo,), indirect=True)
def test_payload(loop, port):
    response = loop.run_until_complete(
        pywreck.post(
            "localhost",
            "/foo",
            headers={"user-agent": "pywreck test, yo!"},
            payload=b"payload, yo!",
            port=port,
            ssl=False,
            timeout=0.2,
        )
    )
    assert response.status == 200
    expected_data = (
        b"POST /foo HTTP/1.1\r\n"
        b"host:localhost\r\n"
        b"user-agent:pywreck test, yo!\r\n"
        b"content-length:12\r\n"
        b"\r\n"
        b"payload, yo!"
    )
    assert response.data == expected_data


@pytest.mark.parametrize(
    "method",
    ("", "get", "post", "put", "delete"),
)
@pytest.mark.parametrize("handler", (handle_echo,), indirect=True)
def test_default_headers_and_payload(loop, port, method):
    if not method:
        method = "GET"
        f = functools.partial(pywreck.request, method)
    else:
        f = getattr(pywreck, method)

    response = loop.run_until_complete(f("localhost", "/", port=port, ssl=False))
    assert response.status == 200
    data = response.data.split(b"\r\n")
    assert len(data) == 5
    assert data[-2:] == [b"", b""]
    assert data[:2] == [
        f"{method.upper()} / HTTP/1.1".encode("latin1"),
        b"host:localhost",
    ]
    assert data[2].startswith(b"user-agent:pywreck/")


@pytest.mark.parametrize("handler", (handle_multi_response_headers,), indirect=True)
def test_multi_line_response_header(loop, port):
    response = loop.run_until_complete(
        pywreck.get("localhost", "/", port=port, ssl=False, timeout=0.2)
    )
    assert response.status == 200
    assert response.headers == {"foo": "1:1,2:2"}
    assert response.data == b""


@pytest.mark.parametrize("handler", (handle_cookies,), indirect=True)
def test_cookies(loop, port):
    response = loop.run_until_complete(
        pywreck.get("localhost", "/", port=port, ssl=False, timeout=0.2)
    )
    assert response.status == 200
    assert response.headers == {"set-cookie": "foo=bar;boo=baz"}


@pytest.mark.parametrize("handler", (handle_chunked,), indirect=True)
def test_chunked(loop, port):
    async def _async():
        connection = await pywreck.Connection.create(
            "localhost",
            port=port,
            ssl=False,
            timeout=None,
        )

        async with connection:
            for _ in range(2):
                response = await connection.request(
                    "GET",
                    "/",
                    headers={"user-agent": "pywreck test, yo!"},
                )

                assert response.status == 200
                assert response.headers == {"transfer-encoding": "chunked"}
                assert response.data == (b"*" * 16 + b"foo")

    loop.run_until_complete(_async())


@pytest.mark.parametrize("handler", (handle_close,), indirect=True)
def test_protocol_error(loop, port):
    with pytest.raises(ConnectionResetError):
        loop.run_until_complete(
            pywreck.get("localhost", "/", port=port, ssl=False, timeout=0.2)
        )



# Copyright 2020 Allan Feldman
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio


async def read_request(reader):
    request = []
    content_length = 0

    while True:
        line = await asyncio.wait_for(
            reader.readline(),
            timeout=0.1,
        )
        request.append(line)
        if not line.rstrip():
            break

        if line.lower().startswith(b"content-length"):
            content_length = int(line[15:])

    data = await asyncio.wait_for(
        reader.readexactly(content_length),
        timeout=0.1,
    )
    request.append(data)

    return b"".join(request)


async def handle_echo(reader, writer):
    while True:
        output = await read_request(reader)
        if not output:
            break
        writer.write(b"HTTP/1.1 200 OK\r\n")
        writer.write(f"content-length: {len(output)}\r\n\r\n".encode("latin1"))
        writer.write(output)
        await writer.drain()

    writer.close()
    await writer.wait_closed()


async def handle_multi_response_headers(reader, writer):
    await read_request(reader)
    writer.write(b"HTTP/1.1 200 OK\r\n")
    writer.write(b"foo:1:1\r\n")
    writer.write(b"foo:2:2\r\n")
    writer.write(b"\r\n")
    await writer.drain()
    writer.close()


async def handle_chunked(reader, writer):
    while True:
        output = await read_request(reader)
        if not output:
            break

        writer.write(b"HTTP/1.1 200 OK\r\n")
        writer.write(b"transfer-encoding: chunked\r\n")
        writer.write(b"\r\n")
        await writer.drain()

        chunks = [b"*" * 16, b"foo", b""]
        for chunk in chunks:
            hex_len = b"%x\r\n" % len(chunk)
            writer.write(hex_len)
            writer.write(chunk + b"\r\n")
            await writer.drain()

    writer.close()
    await writer.wait_closed()


async def handle_cookies(reader, writer):
    await read_request(reader)
    writer.write(b"HTTP/1.1 200 OK\r\n")
    writer.write(b"set-cookie: foo=bar\r\n")
    writer.write(b"set-cookie: boo=baz\r\n")
    writer.write(b"\r\n")
    await writer.drain()
    writer.close()


async def handle_close(reader, writer):
    writer.transport.abort()



# Copyright 2020 Allan Feldman
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import os.path
import ssl
from dataclasses import dataclass
from types import TracebackType
from typing import Dict, Optional, Type, Union

try:
    with open(os.path.join(os.path.dirname(__file__), "version.txt")) as f:
        __version__ = f.read()
except Exception:  # pragma: no cover
    __version__ = "0.0.0"  # pragma: no cover


class _HttpReader:
    def __init__(self, reader: asyncio.StreamReader, timeout: Optional[float]) -> None:
        self.reader = reader
        self.timeout = timeout

    async def readline(self) -> str:
        coro = self.reader.readline()
        result = await asyncio.wait_for(coro, timeout=self.timeout)
        return result.decode("latin1")

    async def readexactly(self, n: int) -> bytes:
        coro = self.reader.readexactly(n)
        return await asyncio.wait_for(coro, timeout=self.timeout)

    async def readuntil(self, separator: bytes) -> bytes:
        coro = self.reader.readuntil(separator)
        return await asyncio.wait_for(coro, timeout=self.timeout)


class _HttpWriter:
    def __init__(self, writer: asyncio.StreamWriter, timeout: Optional[float]) -> None:
        self.writer = writer
        self.timeout = timeout
        self.transport = writer.transport

    def write(self, data: bytes) -> None:
        return self.writer.write(data)

    def write_ascii(self, data: str) -> None:
        return self.writer.write(data.encode("latin1"))

    def close(self) -> None:
        return self.writer.close()

    async def wait_closed(self) -> None:
        coro = self.writer.wait_closed()
        return await asyncio.wait_for(coro, timeout=self.timeout)


@dataclass(frozen=True)
class Response:
    """HTTP Response Container"""

    status: int
    headers: Dict[str, str]
    data: bytes


class Connection:
    """Streaming connection wrapper

    Provides an interface to make HTTP requests via a streaming connection.

    It is not recommended to instantiate :class:`Connection` objects directly;
    use :meth:`Connection.create()` instead.
    """

    def __init__(
        self,
        reader: asyncio.StreamReader,
        writer: asyncio.StreamWriter,
        lock: asyncio.Lock,
        host: str,
        timeout: Optional[float],
    ):
        self._reader = _HttpReader(reader, timeout)
        self._writer = _HttpWriter(writer, timeout)
        self._lock = lock
        self._host = host

    @classmethod
    async def create(
        cls,
        host: str,
        port: int = 443,
        ssl: Union[bool, ssl.SSLContext] = True,
        timeout: Optional[float] = 5.0,
    ) -> "Connection":
        """Create a Connection

        :param host: Host string controlling both the DNS request and the
            host header.
        :type host: str
        :param port: (optional) The TCP port used for the connection. Default: 443
        :type port: int
        :param ssl: (optional) Indicates if SSL is to be used in
            establishing the connection. Also accepts an SSLContext object.
            Default: True
        :type ssl: bool or ssl.SSLContext
        :param timeout: (optional) Timeout in seconds, used when retrieving a response.
            Default: 5 seconds.
        :type timeout: float

        :rtype: Connection
        """
        reader, writer = await asyncio.open_connection(host, port, ssl=ssl)
        return Connection(reader, writer, asyncio.Lock(), host, timeout)

    async def __aenter__(self) -> "Connection":
        return self

    async def request(
        self,
        method: str,
        uri: str,
        payload: bytes = b"",
        headers: Optional[Dict[str, str]] = None,
    ) -> Response:
        """Make an HTTP request

        Drives a full request/response cycle using the HTTP/1.1 protocol.

        No exception handling is done for:

        * Malformed HTTP/1.1 responses
        * Network communication errors

        The raw exceptions will be raised for any violation of HTTP/1.1
        protocol.

        :param method: HTTP method string send in the request line.
        :type method: str
        :param uri: Request-Uniform Resource Identifier (URI), usually
            the absolute path to the resource being requested.
        :type uri: str
        :param payload: (optional) The encoded HTTP request body bytes to be
            sent. Default: b""
        :type payload: bytes
        :param headers: (optional) A dictionary of headers to be sent
            with the request. Default: {}
        :type headers: dict

        :rtype: Response
        """
        # Since the connection can be a shared resource, we must ensure
        # exclusive access for the duration of the request/response cycle
        async with self._lock:
            reader, writer = self._reader, self._writer
            request = f"{method} {uri} HTTP/1.1\r\n"
            writer.write_ascii(request)

            request_headers = {
                "host": self._host,
                "user-agent": f"pywreck/{__version__}",
            }
            if payload:
                request_headers["content-length"] = str(len(payload))

            if headers:
                request_headers.update(headers)

            for header_name, header_value in request_headers.items():
                writer.write_ascii(f"{header_name}:{header_value}\r\n")

            # Finish request metadata section
            writer.write(b"\r\n")

            # Send payload
            writer.write(payload)

            response_line = await reader.readline()
            status = int(response_line.split(" ", 2)[1])

            response_headers: Dict[str, str] = {}
            content_length = 0
            chunked = False

            while True:
                header_line = await reader.readline()
                header_line = header_line.rstrip()
                if not header_line:
                    break

                header_name, header_value = header_line.split(":", 1)
                header_name = header_name.rstrip().lower()
                header_value = header_value.lstrip()

                if header_name in response_headers:
                    separator = "," if header_name != "set-cookie" else ";"
                    response_headers[header_name] += separator + header_value
                else:
                    response_headers[header_name] = header_value

            if method != "HEAD":
                if "content-length" in response_headers:
                    content_length = int(response_headers["content-length"])

                chunked = response_headers.get("transfer-encoding", "") == "chunked"

            if chunked:
                response_chunks = []
                while True:
                    chunk_len_bytes = await reader.readuntil(b"\r\n")
                    content_length = int(chunk_len_bytes.rstrip(), 16)
                    part = await reader.readexactly(content_length + 2)
                    if not content_length:
                        break
                    response_chunks.append(part[:-2])

                response_data = b"".join(response_chunks)
            else:
                response_data = await reader.readexactly(content_length)

        return Response(status, response_headers, response_data)

    async def __aexit__(
        self,
        exc: Optional[Type[BaseException]],
        value: Optional[BaseException],
        tb: Optional[TracebackType],
    ) -> None:
        await self.close()

    async def close(self) -> None:
        """Close the connection"""
        writer = self._writer
        writer.close()
        try:
            await writer.wait_closed()
        except (OSError, asyncio.TimeoutError):
            assert isinstance(writer.transport, asyncio.WriteTransport)
            writer.transport.abort()


async def request(
    method: str,
    host: str,
    uri: str,
    payload: bytes = b"",
    headers: Optional[Dict[str, str]] = None,
    port: int = 443,
    timeout: Optional[float] = 5.0,
    ssl: Union[bool, ssl.SSLContext] = True,
) -> Response:
    """Make a full HTTP request

    Drives a full request/response cycle using the HTTP/1.1 protocol.

    No exception handling is done for:

    * Malformed HTTP/1.1 responses
    * Network communication errors

    The raw exceptions will be raised for any violation of HTTP/1.1
    protocol.

    :param method: HTTP method string send in the request line.
    :type method: str
    :param host: Host string controlling both the DNS request and the
        host header.
    :type host: str
    :param uri: Request-Uniform Resource Identifier (URI), usually
        the absolute path to the resource being requested.
    :type uri: str
    :param payload: The encoded HTTP request body bytes to be sent.
    :type payload: bytes
    :param headers: (optional) A dictionary of headers to be sent
        with the request. Default: {}
    :type headers: dict
    :param port: (optional) The TCP port used for the connection. Default: 443
    :type port: int
    :param timeout: (optional) Timeout in seconds, used when retrieving a response.
        Default: 5 seconds.
    :type timeout: float
    :param ssl: (optional) Indicates if SSL is to be used in
        establishing the connection. Also accepts an SSLContext object.
        Default: True
    :type ssl: bool or ssl.SSLContext

    :rtype: Response
    """
    connection = await Connection.create(host, port, ssl=ssl, timeout=timeout)
    try:
        return await connection.request(
            method=method,
            uri=uri,
            payload=payload,
            headers=headers,
        )
    finally:
        await connection.close()


async def get(
    host: str,
    uri: str,
    payload: bytes = b"",
    headers: Optional[Dict[str, str]] = None,
    port: int = 443,
    timeout: Optional[float] = 5.0,
    ssl: Union[bool, ssl.SSLContext] = True,
) -> Response:
    return await request("GET", host, uri, payload, headers, port, timeout, ssl)


async def head(
    host: str,
    uri: str,
    payload: bytes = b"",
    headers: Optional[Dict[str, str]] = None,
    port: int = 443,
    timeout: Optional[float] = 5.0,
    ssl: Union[bool, ssl.SSLContext] = True,
) -> Response:
    return await request("HEAD", host, uri, payload, headers, port, timeout, ssl)


async def post(
    host: str,
    uri: str,
    payload: bytes = b"",
    headers: Optional[Dict[str, str]] = None,
    port: int = 443,
    timeout: Optional[float] = 5.0,
    ssl: Union[bool, ssl.SSLContext] = True,
) -> Response:
    return await request("POST", host, uri, payload, headers, port, timeout, ssl)


async def put(
    host: str,
    uri: str,
    payload: bytes = b"",
    headers: Optional[Dict[str, str]] = None,
    port: int = 443,
    timeout: Optional[float] = 5.0,
    ssl: Union[bool, ssl.SSLContext] = True,
) -> Response:
    return await request("PUT", host, uri, payload, headers, port, timeout, ssl)


async def delete(
    host: str,
    uri: str,
    payload: bytes = b"",
    headers: Optional[Dict[str, str]] = None,
    port: int = 443,
    timeout: Optional[float] = 5.0,
    ssl: Union[bool, ssl.SSLContext] = True,
) -> Response:
    return await request("DELETE", host, uri, payload, headers, port, timeout, ssl)
